"
 COPYRIGHT (c) 1993 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

'From Smalltalk/X, Version:2.10.5 on 7-may-1995 at 11:49:31 am'!

Object subclass:#WindowSensor
	 instanceVariableNames:'eventSemaphore damage mouseAndKeyboard compressMotionEvents
		ignoreUserInput exposeEventSemaphore catchExpose
		gotExpose gotOtherEvent translateKeyboardEvents shiftDown
		controlDown metaDown altDown eventListener keyboardListener'
	 classVariableNames:'ControlCEnabled EventListener'
	 poolDictionaries:''
	 category:'Interface-Support'
!

WindowSensor comment:'
COPYRIGHT (c) 1993 by Claus Gittinger
	      All Rights Reserved

$Header: /cvs/stx/stx/libview/WindowSensor.st,v 1.24 1995-05-16 17:14:42 claus Exp $
'!

!WindowSensor class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1993 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
!

version
"
$Header: /cvs/stx/stx/libview/WindowSensor.st,v 1.24 1995-05-16 17:14:42 claus Exp $
"
!

documentation
"
    Instances of this class keep track of events and damage areas for a group of 
    views. All incoming expose rectangles and events (from Workstation) are 
    collected here, until someone (usually the windowGroup process)
    gets a chance to handle them. 
    In contrast to ST-80 (which has one windowSensor per window), ST/X usually
    only assigns one sensor per windowGroup.
    (however, you could manually arrange for per view private sensors - at least, theoretically)

    When adding an expose rectangle, WindowSensor tries to merge the rectangle 
    with the list of existing damages to minimize redrawing.

    Beside the above, windowSensors provide facilities (hooks) to allow
    a so-called 'eventListener' to get the events before it is enterred into
    the queue. There are 3 possible listening hooks available:

	a global EventListener - gets keybd/mouse events for all views
	a per-sensor eventListener - gets only keybd/mouse events for this sensors wGroup
	a per-sensor keyboardListener - only gets keyboard events for this sensors wGroup

    (actually, there is a fourth mechanism, event delegation which allows
     delegation of key- and buttonEvents of a specific view)

    The global eventListener is installed via a class method (eventListener:) to
    the WindowSensor class; local listeners are installed via instance methods.
    Each listener should return true, if it handled the event and that event should
    therefore NOT be enqueued. Likewise, if it returns false, the event is
    processed as usual.
    The global listener is called before the local listener, which is called
    before the keyboard listener. If any returns true, later listeners wont get
    the event.
    EventListeners were added to allow the implementation of event recorders
    or other spy functionality.

    Notice, that beside event listening, you can also define a delegate for
    a views keyboard and button events. 
    Read the documentation in WindowEvent for more info.


    instance variables:
	eventSemaphore          <Semaphore>     the semaphore to be signalled when an event
						(or damage) arrives

	damage                  <Collection>    collection of damage events

	mouseAndKeyboard        <Collection>    collection of user events

	compressMotionEvents    <Boolean>       if true, multiple motion events are
						compressed to one event. If false, each
						event is handled individual.
						(should be set to false when doing free-hand drawing)

	ignoreUserInput         <Boolean>       if true, key & button events are ignored
						(usually set to true by WindowGroup, while a
						 modalbox covers a view)

	shiftDown               <Boolean>       true while shift/meta/control-key is pressed
	metaDown                                (to support ST-80 style query: sensor shiftDown)
	controlDown
	altDown                                 (notice, that on most systems, alt and meta key is
						 the same, both reported as #Alt)

	exposeEventSemaphore    <Semaphore>     X-special: semaphore to be signalled when
						expose event arrives after a copyArea.

	catchExpose             <Boolean>       true, while waiting for an expose event 
						(after a copyArea)

	gotExpose               <Boolean>       set to true, when an expose event arrives
						(after a copyarea)

	gotOtherEvent           <Boolean>       set to true if other events arrive while
						waiting for expose (after a copyarea).

	translateKeyboardEvents <Boolean>       if true, keyboard events are translated via
						the devices leyboardMap; if false, they
						are reported as raw-keys. Default is true.

	eventListener           <Object>        if non nil, this one will get all pointer
						and keyboard events for this sensors views first.
						If it returns true, the event is supposed to
						be already handled by the listener and not sent to
						the view. If false, the event is handled as usual.
						This allows applications to catch events for any of
						its views.

	keyboardListener        <Object>        if non nil, this one will get all keyboard events 
						for this sensors views first (but after the eventListener,
						if any).
						If it returns true, the event is supposed to
						be already handled by the listener and not sent to
						the view. If false, the event is handled as usual.
						This allows applications to catch events for any of
						its views.
						ApplicationModels can catch keyboard input with:
						    postOpenWith:aBuilder
							aBuilder window sensor keyboardListener:self


    class variables:

	ControlCEnabled         <Boolean>       if true (which is the default) Control-C
						will interrupt the process handling the
						view.
						For secure stand-alone applications,
						this can be set to false, in which case 
						Control-C does NOT interrupt the process.

	EventListener           <Object>        if non nil, this one will get all pointer
						and keyboard events for ALL views first.
						If it returns true, the view is supposed to
						be handled by the listener. If false, the
						event is handled as usual.
						This allows overall event catchers to be
						installed for example to implement event
						recorders, active help managers etc.
"
! !

!WindowSensor class methodsFor:'initialization'!

initialize
    ControlCEnabled := true.
! !

!WindowSensor class methodsFor:'instance creation'!

new
    ^ self basicNew initialize
! !

!WindowSensor class methodsFor:'accessing'!

controlCEnabled:aBoolean
    "enable/disable Control-C processing. 
     If enabled, pressing CNTL-C in a view will interrupt it and bring
     its process into the debugger (actually raising signal).
     Otherwise, CNTL-C is sent to the view like any other key.
     The default is true (enabled).
    "

    ControlCEnabled := aBoolean
!

eventListener:aListener
    "set the eventListener 
     - see documentation for what this can be used for"

    EventListener := aListener
!

eventListener
    "return the eventListener 
     - see documentation for what this can be used for"

    ^ EventListener
! !

!WindowSensor class methodsFor:'queries'!

cursorPoint
    "ST-80 compatibility: 
     return the position of the cursor on the default display"

    ^ Display pointerPosition

    "
     WindowSensor cursorPoint
    "
! !

!WindowSensor methodsFor:'event processing'!

notifyEventArrival
    "an event arrived - if there is an eventSemaphore,
     signal it, to wake up any windowGroup process"

    catchExpose == true ifTrue:[
	"
	 dont wake up, if we are currently waiting for an expose
	 but remember arrival of something.
	"
	gotOtherEvent := true.
	^ self
    ].
    eventSemaphore notNil ifTrue:[
	eventSemaphore signal
    ]
!

key:key state:onOrOff
    "update the state of the shiftDown/metaDown and controlDown
     flags"

    (key == #Shift
    or:[key == #'Shift_R' 
    or:[key == #'Shift_L']]) ifTrue:[
	shiftDown := onOrOff.
	^ self
    ].
    (key == #Alt
    or:[key == #'Alt_R' or:[key == #'Alt_L']])  ifTrue:[
	altDown := onOrOff.
	^ self
    ].
    (key == #Meta
    or:[key == #'Meta_R' or:[key == #'Meta_L']]) ifTrue:[
	metaDown := onOrOff.
	^ self
    ].
    (key == #Control
    or:[key == #'Control_R' or:[key == #'Control_L']]) ifTrue:[
	controlDown := onOrOff.
	^ self
    ].
!

noExposeView:aView
    "an noexpose event arrived - this is sent from the device (Display)"

    gotExpose := true.
    exposeEventSemaphore notNil ifTrue:[
	exposeEventSemaphore signal
    ]
!

exposeX:left y:top width:width height:height view:aView
    "an expose event arrived - this is sent from the device (Display)"

    (self addDamage:(Rectangle left:left top:top width:width height:height) view:aView) ifTrue:[
	self notifyEventArrival
    ]
!

buttonMotion:state x:x y:y view:aView
    "mouse was moved - this is sent from the device (Display)"

    |args|

    EventListener notNil ifTrue:[
	(EventListener buttonMotion:state x:x y:y view:aView) ifTrue:[^ self]
    ].
    eventListener notNil ifTrue:[
	(eventListener buttonMotion:state x:x y:y view:aView) ifTrue:[^ self]
    ].

    ignoreUserInput == true ifTrue:[
	^ self
    ].
    args := Array with:state with:x with:y.

    compressMotionEvents ifTrue:[
	"
	 merge with last motion
	"
	mouseAndKeyboard reverseDo:[:ev |
	    ev notNil ifTrue:[
		((ev type == #buttonMotion:x:y:) 
		and:[(ev view == aView)
		and:[(ev arguments at:1) == state]]) ifTrue:[
		    ev arguments:args.
		    ^ self
		]
	    ]
	]
    ].
    mouseAndKeyboard
	addLast:(WindowEvent
		     for:aView
		     type:#buttonMotion:x:y:
		     arguments:args).
    self notifyEventArrival
!

pointerLeave:state view:aView
    "mouse cursor was moved out of the view - this is sent from the device (Display)"

    EventListener notNil ifTrue:[
	(EventListener pointerLeave:state view:aView) ifTrue:[^ self]
    ].
    eventListener notNil ifTrue:[
	(eventListener pointerLeave:state view:aView) ifTrue:[^ self]
    ].

    mouseAndKeyboard
	 addLast:(WindowEvent
		     for:aView
		     type:#pointerLeave: 
		     arguments:(Array with:state)).
    self notifyEventArrival
!

buttonRelease:button x:x y:y view:aView
    "mouse button was released- this is sent from the device (Display)"

    EventListener notNil ifTrue:[
	(EventListener buttonRelease:button x:x y:y view:aView) ifTrue:[^ self]
    ].
    eventListener notNil ifTrue:[
	(eventListener buttonRelease:button x:x y:y view:aView) ifTrue:[^ self]
    ].

    ignoreUserInput == true ifTrue:[
	^ self
    ].
    mouseAndKeyboard
	addLast:(WindowEvent
		     for:aView 
		     type:#buttonRelease:x:y:
		     arguments:(Array with:button with:x with:y)).
    self notifyEventArrival
!

buttonPress:button x:x y:y view:aView
    "mouse button was pressed - this is sent from the device (Display)"

    EventListener notNil ifTrue:[
	(EventListener buttonPress:button x:x y:y view:aView) ifTrue:[^ self]
    ].
    eventListener notNil ifTrue:[
	(eventListener buttonPress:button x:x y:y view:aView) ifTrue:[^ self]
    ].

    ignoreUserInput == true ifTrue:[
	^ self
    ].
    mouseAndKeyboard 
	addLast:(WindowEvent 
		     for:aView
		     type:#buttonPress:x:y:
		     arguments:(Array with:button with:x with:y)).
    self notifyEventArrival
!

graphicExposeX:left y:top width:width height:height view:aView
    "a graphic expose event arrived - this is sent from the device (Display)"

    self addDamage:(left @ top extent:width @ height) view:aView.
!

focusInView:aView
    "view got input focus - this is sent from the device (Display)"

    mouseAndKeyboard
	 addLast:(WindowEvent
		     for:aView
		     type:#focusIn).
    self notifyEventArrival
!

pointerEnter:state x:x y:y view:aView
    "mouse cursor was moved into the view - this is sent from the device (Display)"

    EventListener notNil ifTrue:[
	(EventListener pointerEnter:state x:x y:y view:aView) ifTrue:[^ self]
    ].
    eventListener notNil ifTrue:[
	(eventListener pointerEnter:state x:x y:y view:aView) ifTrue:[^ self]
    ].

    mouseAndKeyboard
	addLast:(WindowEvent
		  for:aView
		  type:#pointerEnter:x:y:
		  arguments:(Array with:state with:x with:y)).
    self notifyEventArrival
!

buttonMultiPress:button x:x y:y view:aView
    "mouse button was pressed - this is sent from the device (Display)"

    EventListener notNil ifTrue:[
	(EventListener buttonMultiPress:button x:x y:y view:aView) ifTrue:[^ self]
    ].
    eventListener notNil ifTrue:[
	(eventListener buttonMultiPress:button x:x y:y view:aView) ifTrue:[^ self]
    ].

    ignoreUserInput == true ifTrue:[
	^ self
    ].
    mouseAndKeyboard
	 addLast:(WindowEvent
		      for:aView
		      type:#buttonMultiPress:x:y:
		      arguments:(Array with:button with:x with:y)).
    self notifyEventArrival
!

mappedView:aView
    "view was mapped (from window manager) - this is sent from the device (Display)"

    damage
	 addLast:(WindowEvent
		     for:aView
		     type:#mapped).
    self notifyEventArrival
!

keyPress:key x:x y:y view:aView
    "key was pressed - this is sent from the device (Display).
     beside the keyboard translation, CntlC processing is done here."

    |xlatedKey group process|

    self key:key state:true. 

    EventListener notNil ifTrue:[
	(EventListener keyPress:key x:x y:y view:aView) ifTrue:[^ self]
    ].
    eventListener notNil ifTrue:[
	(eventListener keyPress:key x:x y:y view:aView) ifTrue:[^ self]
    ].
    keyboardListener notNil ifTrue:[
	(keyboardListener keyPress:key x:x y:y view:aView) ifTrue:[^ self]
    ].

    translateKeyboardEvents ifTrue:[
	xlatedKey := aView device translateKey:key.
    ] ifFalse:[
	xlatedKey := key.
    ].
    xlatedKey isNil ifTrue:[^ self].

    (xlatedKey == #CtrlV) ifTrue:[
	'Smalltalk/X ' errorPrint. 
	Smalltalk versionString errorPrint. ' of ' errorPrint.
	Smalltalk versionDate errorPrintNL.
	Smalltalk copyrightString errorPrintNL.
    ].

    ((xlatedKey == #UserInterrupt) and:[ControlCEnabled]) ifTrue:[
	"
	 Special handling for Ctrl-C: interrupt the underlying process.

	 cannot halt here (this would stop the event-dispatcher),
	 but instead interrupt the underlying process and have it
	 perform the userInterrupt in the interrupt-method.
	"
	group := aView windowGroup.
	group notNil ifTrue:[
	    process := group process.
	    process notNil ifTrue:[
		process interruptWith:[process userInterrupt]
	    ]
	].
	^ self
    ].

    ignoreUserInput == true ifTrue:[
	^ self
    ].
    mouseAndKeyboard
	addLast:(WindowEvent
		     for:aView
		     type:#keyPress:x:y:
		     arguments:(Array with:xlatedKey with:x with:y)).
    self notifyEventArrival
!

coveredBy:sibling view:aView
    "aView was covered by one of its siblings - this is sent from the device (Display)"

    damage
	 addLast:(WindowEvent
		     for:aView
		     type:#coveredBy:
		     arguments:(Array with:sibling)).
    self notifyEventArrival
!

configureX:x y:y width:w height:h view:aView
    "a views size or position has changed - this is sent from the device (Display)"

    damage
	addLast:(WindowEvent
		     for:aView
		     type:#configureX:y:width:height:
		     arguments:(Array with:x with:y with:w with:h)).
    self notifyEventArrival
!

unmappedView:aView
    "view was unmapped (from window manager) - this is sent from the device (Display)"

    damage
	 addLast:(WindowEvent
		     for:aView
		     type:#unmapped).
    self notifyEventArrival
!

focusOutView:aView
    "view lost input focus - this is sent from the device (Display)"

    mouseAndKeyboard
	 addLast:(WindowEvent
		     for:aView
		     type:#focusOut).
    self notifyEventArrival
!

buttonShiftPress:button x:x y:y view:aView
    "mouse button was pressed - this is sent from the device (Display)"

    EventListener notNil ifTrue:[
	(EventListener buttonShiftPress:button x:x y:y view:aView) ifTrue:[^ self]
    ].
    eventListener notNil ifTrue:[
	(eventListener buttonShiftPress:button x:x y:y view:aView) ifTrue:[^ self]
    ].

    ignoreUserInput == true ifTrue:[
	^ self
    ].
    mouseAndKeyboard
	addLast:(WindowEvent
		     for:aView
		     type:#buttonShiftPress:x:y:
		     arguments:(Array with:button with:x with:y)).
    self notifyEventArrival
!

keyRelease:key x:x y:y view:aView
    "key was released - this is sent from the device (Display)."

    |xlatedKey|

    self key:key state:false. 

    EventListener notNil ifTrue:[
	(EventListener keyRelease:key x:x y:y view:aView) ifTrue:[^ self]
    ].
    eventListener notNil ifTrue:[
	(eventListener keyRelease:key x:x y:y view:aView) ifTrue:[^ self]
    ].
    keyboardListener notNil ifTrue:[
	(keyboardListener keyRelease:key x:x y:y view:aView) ifTrue:[^ self]
    ].

    ignoreUserInput == true ifTrue:[
	^ self
    ].
    translateKeyboardEvents ifTrue:[
	xlatedKey := aView device translateKey:key.
    ] ifFalse:[
	xlatedKey := key.
    ].
    xlatedKey isNil ifTrue:[^ self].

    mouseAndKeyboard
	addLast:(WindowEvent
		     for:aView
		     type:#keyRelease:x:y:
		     arguments:(Array with:xlatedKey with:x with:y)).
    self notifyEventArrival
!

terminateView:aView
    "view should terminate (from window manager) - this is sent from the device (Display)"

    self flushEventsFor:aView.
    damage
	 addLast:(WindowEvent
		     for:aView
		     type:#terminate).
    self notifyEventArrival
!

saveAndTerminateView:aView
    "view should save & terminate (from window manager) - this is sent from the device (Display)"

    self flushEventsFor:aView.
    damage
	 addLast:(WindowEvent
		     for:aView
		     type:#saveAndTerminate).
    self notifyEventArrival
!

destroyedView:aView
    "view was destroyed (from window manager) - this is sent from the device (Display)"

    "at this time, the view is already gone; remove
     all pending events for this one ..."

    self flushEventsFor:aView.
    damage
	 addLast:(WindowEvent
		     for:aView
		     type:#destroyed).
    self notifyEventArrival
! !

!WindowSensor methodsFor:'event queue'!

nextDamage
    "retrieve the next damage rectangle or nil, if there is none.
     Remove it from the queue."

    |d wasBlocked|

    [d isNil] whileTrue:[
	damage size == 0 ifTrue:[^ nil].
	"
	 be careful: events are inserted at higher prio ...
	"
	wasBlocked := OperatingSystem blockInterrupts.
	d := damage removeFirst.
	wasBlocked ifFalse:[OperatingSystem unblockInterrupts].
    ].
    ^ d
!

nextEvent
    "retrieve the next event or nil, if there is none.
     Remove it from the queue."

    |e wasBlocked|

    [e isNil] whileTrue:[
	mouseAndKeyboard size == 0 ifTrue:[^ nil].
	"
	 be careful: events are inserted at higher prio ...
	"
	wasBlocked := OperatingSystem blockInterrupts.
	e := mouseAndKeyboard removeFirst.
	wasBlocked ifFalse:[OperatingSystem unblockInterrupts].
    ].
    ^ e
!

addDamage:aRectangle view:aView
    "Add aRectangle to the damage list.
     Try to merge incoming rectangles with the existing damage rectangles.
     Incoming rectangles which are completely contained in any existing damage rect are ignored,
     any existing damage rectangle which is completely contained in the incoming rectangle
     is replaced.
     There could be much more optimization here (for example joining rectangles) ...
     Except for special cases (moveOpaque of a view over one of my views),
     these optimizations are not noticable."

    |count temp index newEvent 
     sz "{ Class: SmallInteger }" |

    sz := damage size.
    sz == 0 ifTrue: [
	newEvent := WindowEvent damageFor:aView rectangle:aRectangle.
	damage := OrderedCollection with:newEvent.
	^ true
    ].

    "
     first look, if this rectangle is already in the expose list;
     if so, dont add to queue
    "
    damage do: [:aDamage | 
	aDamage notNil ifTrue:[
	    aDamage isDamage ifTrue:[
		aDamage view == aView ifTrue:[
		    ((aDamage rectangle) contains:aRectangle) ifTrue: [^false]
		]
	    ]
	].
    ].

    "
     then look, if new rectangle contains any in the expose list;
     if so, remove the old damage (here, by nilling it in the queue)
    "
    count := 0.
    1 to:sz do:[:i |
	|aDamage|

	aDamage := damage at:i.
	aDamage notNil ifTrue:[
	    aDamage isDamage ifTrue:[
		(aDamage view) == aView ifTrue:[
		    (aRectangle contains:(aDamage rectangle)) ifTrue: [ 
			damage at:i put:nil.
			count := count + 1
		    ]
		]
	    ]
	]
    ].

    "
     if we nilled more then 10 events, reorganize the queue
     (this avoids excessive reorganization of the input queue)
    "
    count > 10 ifTrue: [
	temp := OrderedCollection new:(sz - count + 1).
	index := 1.
	damage do:[:aDamage |
	    aDamage notNil ifTrue: [
		temp add: aDamage.
	    ]
	].
	damage := temp
    ].
    newEvent := WindowEvent damageFor:aView rectangle:aRectangle.
    damage add:newEvent.
    ^ true
!

pendingEvent
    "retrieve the next pending event or nil, if there is none.
     Do not remove it from the queue."

    |e wasBlocked|

    [e isNil] whileTrue:[
	mouseAndKeyboard size == 0 ifTrue:[^ nil].
	e := mouseAndKeyboard first.
	e isNil ifTrue:[
	    wasBlocked := OperatingSystem blockInterrupts.
	    mouseAndKeyboard removeFirst
	    wasBlocked ifFalse:[OperatingSystem unblockInterrupts].
	]
    ].
    ^ e
! !

!WindowSensor methodsFor:'queries '!

hasEvents 
    "return true, if any mouse/keyboard events are pending"

    ^ mouseAndKeyboard size ~~ 0
!

hasDamage 
    "return true, if any exposure events are pending"

    ^ damage size ~~ 0
!

hasDamageFor:aView 
    "return true, if any exposure events are pending for aView"

    damage size ~~ 0 ifTrue:[
	damage do:[:aDamage |
	    aDamage notNil ifTrue:[
		aDamage view == aView ifTrue:[^ true].
	    ].
	]
    ].
    ^ false
!

hasButtonMotionEventsFor:aView 
    "return true, if any buttonMotion events are pending"

    mouseAndKeyboard size ~~ 0 ifTrue:[
	mouseAndKeyboard do:[:anEvent |
	    anEvent notNil ifTrue:[
		anEvent view == aView ifTrue:[
		    anEvent type == #buttonMotion:x:y: ifTrue:[^ true].
		]
	    ].
	]
    ].
    ^ aView device eventPending:#buttonMotion for:aView id
!

eventPending
    "return true, if either damage or events are pending"

    mouseAndKeyboard size ~~ 0 ifTrue:[^ true].
    ^ damage size ~~ 0
!

mousePoint
    "ST-80 compatibility: 
     return the position of the mouse pointer on the default display"

     ^ self class cursorPoint
!

globalOrigin 
    "ST-80 compatibility: 
     dont know what we should return here ...
     ... at least the PD program which uses it works when we return 0@0."

     ^ 0@0
!

shiftDown
    "return true, if any shift key is currently pressed."

    ^ shiftDown
!

controlDown
    "return true, if any CTRL key is currently pressed."

    ^ controlDown
!

metaDown
    "return true, if the meta key is currently pressed.
     Notice, that most keyboards dont have a meta key;
     it is better to use 'sensor metaDown or:[sensor altDown]'."

    ^ metaDown
!

altDown
    "return true, if the meta key is currently pressed.
     Notice, that some keyboards dont have an alt key;
     it is better to use 'sensor metaDown or:[sensor altDown]'."

    ^ altDown
! !

!WindowSensor methodsFor:'special'!

catchExpose
    "start catching noExpose events (must be done BEFORE a bitblt,
     to prepare for the exposeEventSemaphore to be signalled when 
     the noExpose event arrives)."

    gotExpose := false.
    gotOtherEvent := false.
    catchExpose := true.
    exposeEventSemaphore := Semaphore new.
!

waitForExposeFor:aView
    "wait until a noExpose arrives (after a bitblt)"

    Processor activeProcessIsSystemProcess ifTrue:[
	"
	 cannot really suspend, if its a systemProcess
	"
	[gotExpose] whileFalse:[
	    aView device dispatchExposeEventFor:aView id.
	    Processor yield.
	]
    ] ifFalse:[
	[gotExpose] whileFalse:[
	    "
	     just in case we have a (network or software) problem ...
	     explanation: it may happen, that an expose event is totally
	     lost - for example, if the network breaks down.
	     To not block forever, we wait with a timeout, to get out of here
	     if the event does not arrive after 15 seconds.
	    "
	    (exposeEventSemaphore waitWithTimeout:5) ifFalse:[
		aView device synchronizeOutput.
		'WSENSOR: late expose event' errorPrintNL.
		(exposeEventSemaphore waitWithTimeout:10) ifFalse:[
		    'WSENSOR: lost expose event again - ignore' errorPrintNL.
		].
		"
		 you can put a comment around the following line, if you dont
		 like the message ...
		"
		gotExpose := true.
		^ self
	    ]
	].

	"
	 other incoming events have been ignored during the wait.
	 Now handle those ...
	"
	gotOtherEvent ifTrue:[
	    eventSemaphore signal
	].
    ].
    catchExpose := false
!

flushUserEvents
    "throw away all pending user events"

    (mouseAndKeyboard isNil or:[mouseAndKeyboard size > 0]) ifTrue:[
	mouseAndKeyboard := OrderedCollection new
    ].
!

flushExposeEventsFor:aView
    "throw away all pending expose events for aView; 
     this can be done after a full redraw 
     (or in views, which are always doing full redraws -
      instead of drawing the clip-area only)"

    damage notNil ifTrue:[
	1 to:damage size do:[:index |
	    |aDamage|

	    aDamage := damage at:index.
	    aDamage notNil ifTrue:[
		aDamage view == aView ifTrue:[
		    damage at:index put:nil
		]
	    ]
	]
    ].
!

flushUserEventsFor:aView
    "throw away all pending user events for aView" 

    mouseAndKeyboard notNil ifTrue:[
	1 to:mouseAndKeyboard size do:[:i |
	    |anEvent|

	    anEvent := mouseAndKeyboard at:i.
	    anEvent notNil ifTrue:[
		anEvent view == aView ifTrue:[
		    mouseAndKeyboard at:i put:nil
		]
	    ]
	]
    ].
!

flushEventsFor:aView
    "throw away all events for aView"

    self flushExposeEventsFor:aView.
    self flushUserEventsFor:aView.
!

compressKeyPressEventsWithKey:aKey
    "count and remove multiple pending keyPress events for the
     same key, aKey. This is currently used in TextViews to compress
     multiple cursorUp/cursorDown events and do the scroll in one
     operation. (to avoid run-after-cursor on slow displays)"

    |n ev|

    n := 0.
    ev := self pendingEvent.
    [ev notNil and:[ev isKeyPressEvent]] whileTrue:[
	((ev arguments at:1) == aKey) ifTrue:[
	    n := n + 1.
	    self nextEvent.
	    ev := self pendingEvent.
	] ifFalse:[
	    ev := nil
	]
    ].
    ^ n
!

flushExposeEvents
    "throw away all pending expose events; this
     can be done after a full redraw (or in views, which are
     doing full redraws anly)"

    (damage isNil or:[damage size > 0]) ifTrue:[
	damage := OrderedCollection new
    ].
!

pushUserEvent:anEvent
    "manually put an event into the queue - this allows
     simulation of events (implementation of recorders & playback)."

    mouseAndKeyboard addLast:anEvent.
    self notifyEventArrival
!

pushUserEvent:aSelector for:aView withArguments:arguments
    "manually put an event into the queue - this allows
     simulation of events (implementation of recorders & playback)."

    self pushEvent:(WindowEvent 
		     for:aView
		     type:aSelector
		     arguments:arguments).

    "
     |b|
     b := Button label:'test'.
     b open.
     (Delay forSeconds:5) wait.
     b sensor pushEvent:#pointerEnter:x:y: for:b withArguments:#(0 1 1).
     (Delay forSeconds:1) wait.
     b sensor pushEvent:#buttonPress:x:y: for:b withArguments:#(1 1 1).
     (Delay forSeconds:2) wait.
     b sensor pushEvent:#buttonRelease:x:y: for:b withArguments:#(1 1 1).
     (Delay forSeconds:1) wait.
     b sensor pushEvent:#pointerLeave: for:b withArguments:#(0).
    "
!

forwardKeyEventsTo:aView
    "remove all keyboard events and send them to aSensor instead"

"/ 'fwd' printNL.
    1 to:mouseAndKeyboard size do:[:i |
	|anEvent|

	anEvent := mouseAndKeyboard at:i.
	anEvent notNil ifTrue:[
	    anEvent isKeyEvent ifTrue:[
		anEvent view:aView.
		aView sensor pushUserEvent:anEvent.
"/ anEvent type printNL.
		mouseAndKeyboard at:i put:nil
	    ]
	]
    ].
! !

!WindowSensor methodsFor:'accessing'!

eventSemaphore
    "return the semaphore used to signal event arrival"

    ^ eventSemaphore
!

eventSemaphore:aSemaphore
    "set the semaphore used to signal event arrival"

    eventSemaphore := aSemaphore
!

ignoreUserInput:aBoolean
    "turn on/off ignoring of Ctrl-C processing"

    ignoreUserInput := aBoolean
!

ignoreUserInput
    "return true, if Ctrl-C processing is currently turned off"

    ^ ignoreUserInput
!

compressMotionEvents:aBoolean
    "turn on/off motion event compression"

    compressMotionEvents := aBoolean
!

eventListener:aListener
    "set the eventListener 
     - see documentation for what this can be used for"

    eventListener := aListener

!

eventListener
    "return the eventListener 
     - see documentation for what this can be used for"

    ^ eventListener
!

keyboardListener
    "return the keyboardListener 
     - see documentation for what this can be used for"

    ^ keyboardListener
!

keyboardListener:aListener
    "set the keyboardListener 
     - see documentation for what this can be used for"

    keyboardListener := aListener

! !

!WindowSensor methodsFor:'initialization'!

initialize
    "initialize the event queues to empty"

    damage := OrderedCollection new.
    mouseAndKeyboard := OrderedCollection new.
    gotExpose := true.
    catchExpose := false.

    compressMotionEvents := translateKeyboardEvents := true.
    ignoreUserInput := false.
    shiftDown := controlDown := altDown := metaDown := false
!

reinitialize
    "called when an image is restarted;
     reinitialize the event queues to empty; leave other setup as-is"

    self flushUserEvents.
    self flushExposeEvents.
    gotExpose := true.
    catchExpose := false.
! !

WindowSensor initialize!
